import logging

from django.conf import settings
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import Group, User
from django.db import transaction

from .models import KeycloakUser

# Custom authentication backend for Keycloak integration, open to other customization
logger = logging.getLogger(__name__)


class KeycloakBackend(BaseBackend):

    @transaction.atomic
    def authenticate(self, request, user_info=None):
        """
        Authenticate a user based on Keycloak-provided user information.

        Parameters:
            request: The HTTP request object (optional, can be None).
            user_info: A dictionary containing user data from Keycloak.

        Returns:
            A Django User object if authentication is successful, or None otherwise.
        """
        if not user_info:
            logger.warning("Authentication failed: No user_info provided.")
            return None

        # Extract user data from the Keycloak-provided dictionary
        sub = user_info.get("sub")
        username = user_info.get("username")
        first_name = user_info.get("firstName")
        last_name = user_info.get("lastName")
        email = user_info.get("email")

        try:
            roles = (
                user_info.get("resource_access")
                .get(settings.KC_CLIENT_ID)
                .get("roles", [])
            )
        except:
            roles = []

        if not username:
            logger.warning("Authentication failed: Missing username in user_info.")
            return None

        try:
            user, created = User.objects.get_or_create(username=username)
            if created:
                logger.info("Created new user: %s", username)
                user.set_unusable_password()
            user.first_name = first_name or ""
            user.last_name = last_name or ""
            user.email = email or ""
            user.is_staff = "employees" in roles
            user.is_superuser = "admins" in roles

            kc_roles = settings.KC_ROLES
            current_groups = [group.name for group in user.groups.all()]
            current_groups = [
                group for group in current_groups if group not in kc_roles
            ]
            roles = current_groups + roles

            required_groups = [
                Group.objects.get_or_create(name=group_name)[0] for group_name in roles
            ]

            user.groups.set(required_groups)
            user.save()

            KeycloakUser.objects.get_or_create(sub=sub, user=user)

            user.backend = "django_kc_auth.backends.KeycloakBackend"
            return user

        except Exception as e:
            logger.error(
                "Error during authentication for user %s: %s", username, str(e)
            )
            return None

    def get_user(self, user_id):
        """
        Retrieve a User object by its ID.

        Parameters:
            user_id: The primary key of the User object to retrieve.

        Returns:
            A User object if found, or None otherwise.
        """
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



from django.apps import AppConfig


class KeycloakAuthConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_kc_auth"

    def ready(self):
        import logging

        import jwt
        from django.conf import settings

        from .keycloak_openid_config import keycloak_openid

        logger = logging.getLogger(__name__)

        try:
            keys = keycloak_openid.certs()["keys"]
            jwt_dict = next((key for key in keys if key.get("use") == "sig"), None)
            verifying_key = jwt.jwk_from_dict(jwt_dict)
            settings.KC_VERIFYING_KEY = verifying_key
            logger.info("Keycloak verifying key set up.")
        except Exception:
            logger.error("Initialization failed.", exc_info=True)



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("sessions", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="KeycloakUser",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "sub",
                    models.UUIDField(db_index=True, primary_key=True, serialize=False),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="KeycloakSession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "keycloak_session_id",
                    models.UUIDField(db_index=True, editable=False),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "django_session",
                    models.OneToOneField(
                        editable="False",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="sessions.session",
                    ),
                ),
                (
                    "keycloak_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_kc_auth.keycloakuser",
                    ),
                ),
            ],
        ),
    ]



import datetime
import logging
import re

from django.conf import settings
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.http import urlsafe_base64_encode

from .keycloak_openid_config import keycloak_openid

logger = logging.getLogger(__name__)


class AutoKeycloakLoginMiddleware:
    """Middleware that tries to auto log in user if they have
    active Keycloak session in browser
    """

    ignored_routes = [
        f"{reverse('kc_auth_logout-listener')}*",
        f"{reverse('kc_auth_callback')}*",
        "/static/*",
        "/favicon.ico",
        "/__reload__/events/*",
    ] + getattr(settings, "KC_SILENT_LOGOUT_IGNORED_ROUTES", [])

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        login_attempts = request.session.get("kc_login_attempts", 0)
        last_attempt = request.session.get("kc_last_attempt_at")
        now = datetime.datetime.now()
        last_attempt_valid = True
        if last_attempt:
            last_attempt = datetime.datetime.strptime(last_attempt, "%Y-%m-%d %H:%M:%S")
            diff = now - last_attempt
            if diff.total_seconds() < getattr(
                settings, "KC_SILENT_LOGIN_TIMEOUT_SECONDS", 3
            ):
                last_attempt_valid = False

        if (
            request.method == "GET"
            and not request.user.is_authenticated
            and not self.is_ignored(request.path)
            and not request.session.get("soft_logout", None)
            and last_attempt_valid
            and login_attempts
            < getattr(settings, "KC_SILENT_LOGIN_ALLOWED_ATTEMPTS", 5)
        ):
            logger.info("Silent authentication attempt.")

            server_url = request.build_absolute_uri("/")[:-1]
            callback_url = f"{server_url}{reverse('kc_auth_callback')}"

            auth_url = (
                keycloak_openid.auth_url(
                    redirect_uri=callback_url,
                    scope="openid email",
                    state=urlsafe_base64_encode(
                        request.get_full_path().encode("utf-8")
                    ),
                )
                + "&prompt=none"
            )
            request.session["kc_login_attempts"] = login_attempts + 1
            request.session["kc_last_attempt_at"] = now.strftime("%Y-%m-%d %H:%M:%S")
            response = redirect(auth_url)
            return response

        return self.get_response(request)

    def is_ignored(self, path):
        for pattern in self.ignored_routes:
            if re.search(pattern, path):
                return True
        return False
